from django.conf import (
    settings,
)
from django.test.testcases import (
    TestCase,
)
from rest_framework.permissions import (
    IsAuthenticated,
)

from m3_gar_client.backends.m3_rest_gar.proxy_rest.authentication import (
    CsrfExemptSessionAuthentication,
)
from m3_gar_client.backends.m3_rest_gar.proxy_rest.views import (
    get_authentication_classes,
    get_permission_classes,
)


class DefaultsTestCase(TestCase):

    """Проверка конфигурации по-умолчанию.

    Конфигурация соответствует поведению монолитных продуктов.
    """

    def test_defaults(self):
        # Конфиг не задан
        self.assertIsNone(settings.GAR.get('REST'))

        self.assertEqual(
            get_authentication_classes(),
            [CsrfExemptSessionAuthentication]
        )

        self.assertEqual(
            get_permission_classes(),
            [IsAuthenticated]
        )



from abc import (
    ABC,
)
from typing import (
    List,
    Type,
)

from django.conf import (
    settings,
)
from django.utils.module_loading import (
    import_string,
)
from rest_framework.authentication import (
    BaseAuthentication,
)
from rest_framework.permissions import (
    BasePermission,
    IsAuthenticated,
)
from rest_framework.response import (
    Response,
)
from rest_framework.views import (
    APIView,
)

from m3_gar_client.backends.m3_rest_gar.utils import (
    LoaderBase,
    PlaceLoader,
    UIApartmentLoader,
    UIHouseLoader,
    UIHouseOrSteadLoader,
    UIStreetLoader,
)
from m3_gar_client.utils import (
    correct_keyboard_layout,
)

from .authentication import (
    CsrfExemptSessionAuthentication,
)


def __get_config_or_defaults(name, defaults):
    cfg = settings.GAR.get('REST', {}).get(name)
    if cfg is not None:
        return [import_string(path) for path in cfg]
    return defaults


def get_authentication_classes():
    return __get_config_or_defaults(
        'AUTHENTICATION_CLASSES', [CsrfExemptSessionAuthentication]
    )


def get_permission_classes():
    return __get_config_or_defaults(
        'PERMISSION_CLASSES', [IsAuthenticated]
    )


class AbstractSearchView(APIView, ABC):

    authentication_classes: List[Type[BaseAuthentication]] = (
        get_authentication_classes()
    )
    permission_classes: List[Type[BasePermission]] = (
        get_permission_classes()
    )

    # Признак того, что необходимо корректировать пользовательский ввод
    correct_keyboard_input = True
    loader_cls: Type[LoaderBase]

    def post(self, request):
        result = self._get_loader(request).load()

        return self._build_response(
            data=result.rows,
            total=result.total,
        )

    def _get_loader(self, request):
        """Возвращает инстанс загрузчика данных."""
        args, kwargs = self._get_loader_params(request)
        return self.loader_cls(*args, **kwargs)

    def _get_loader_params(self, request):
        filter_ = request.POST.get('filter', '')

        if self.correct_keyboard_input and filter_:
            filter_ = correct_keyboard_layout(filter_)

        return (filter_, ), {}

    def _build_response(self, data, total):
        return Response({
            'rows': data,
            'total': total,
        })


class ParentSearchMixin(AbstractSearchView):
    def _get_loader_params(self, request):
        args, kwargs = super()._get_loader_params(request)
        kwargs.update(parent_id=request.POST['parent'])
        return args, kwargs


class PlaceSearchView(AbstractSearchView):

    loader_cls = PlaceLoader


class StreetSearchView(ParentSearchMixin, AbstractSearchView):

    loader_cls = UIStreetLoader


class HouseSearchView(ParentSearchMixin, AbstractSearchView):

    correct_keyboard_input = False
    loader_cls = UIHouseLoader


class HouseOrSteadSearchView(ParentSearchMixin, AbstractSearchView):

    correct_keyboard_input = False
    loader_cls = UIHouseOrSteadLoader

    def post(self, request):
        result = self._get_loader(request).load()

        return self._build_response(
            data=result.rows,
            total=result.total,
        )


class ApartmentSearchView(ParentSearchMixin, AbstractSearchView):
    correct_keyboard_input = False
    loader_cls = UIApartmentLoader



# Использовать постраничный вывод значений в выпадающем списке
DEFAULT_FIELD_IS_PAGINATED = True

# Минимальное кол-во символов, которое необходимо ввести в поле формы для начала
# отправки запроса для получения данных в выпадающем списке
DEFAULT_FIELD_MIN_CHARS = 2

# Задержка в миллисекундах между завершением набора текста в поле формы
# и отправкой запроса на получение данных для выпадающего списка
DEFAULT_FIELD_QUERY_DELAY_MS = 1500



from abc import (
    ABC,
)
from importlib import (
    import_module,
)

import django
from django.conf import (
    settings,
)

# нужно для импорта констант отсюда в barsdoc
from m3_gar_constants import *

from m3_gar_client.utils import (
    cached_property,
)


if django.VERSION < (3, 2):
    default_app_config = f'{__package__}.apps.AppConfig'


class Config(ABC):
    """Базовый класс для конфигурации пакета."""

    @cached_property
    def backend(self):
        """Бэкенд для доступа к данным ГАР.

        :rtype: :class:`m3_gar_client.backends.base.BackendBase`
        """
        backend_class = import_module(settings.GAR['BACKEND']).Backend
        return backend_class()


#: Конфигурация приложения ``m3_gar_client``.
#:
#: Заполняется экземпляром класса :class:`m3_gar_client.Config`, либо его потомком,
#: при инициализации проекта *до* инициализации приложения ``m3-gar-client``.
config = None


if settings.DEBUG:
    import os

    os.environ['OAUTHLIB_INSECURE_TRANSPORT'] = '1'



"""Бэкенд, проксирующий запросы через веб-приложение."""
from django.urls import (
    path,
)
from django.urls.base import (
    reverse,
)
from django.urls.conf import (
    include,
)
from django.utils.module_loading import (
    import_string,
)

from m3_gar_client.backends.m3_rest_gar.base import (
    BackendBase,
)

from . import (
    const,
)


class Backend(BackendBase):

    """REST-бэкенд для работы с сервером m3-rest-gar.

    .. code::python

      GAR = dict(
          BACKEND='m3_gar_client.backends.m3_rest_gar.proxy_rest',
          ...
      )

    """

    def init(self) -> None:
        from django.conf import (
            settings,
        )

        from m3_gar_client.utils import (
            reload_urlconf,
        )

        from .urls import (
            urlpatterns,
        )

        urlconf = import_string(f'{settings.ROOT_URLCONF}.urlpatterns')
        urlconf += [
            path(
                r'gar/',
                include('m3_gar_client.backends.m3_rest_gar.proxy_rest.urls')
            )
        ]

        reload_urlconf()

    @property
    def place_search_url(self) -> str:
        """URL для поиска населенных пунктов."""

        return reverse(const.SEARCH_PLACE_NAME)

    @property
    def street_search_url(self) -> str:
        """URL для поиска улиц."""

        return reverse(const.SEARCH_STREET_NAME)

    @property
    def house_search_url(self) -> str:
        """URL для запроса списка домов."""

        return reverse(const.SEARCH_HOUSE_OR_STEAD_NAME)

    @property
    def apartment_search_url(self) -> str:
        return reverse(const.SEARCH_APARTMENT)



"""Бэкенд, проксирующий запросы через веб-приложение."""
from uuid import (
    UUID,
)

from m3_gar_client.backends.m3_rest_gar.base import (
    BackendBase,
)


class Backend(BackendBase):
    """Бэкенд для работы с сервером m3-rest-gar."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._pack = None

    @staticmethod
    def _register_parsers():
        """Регистрация парсеров для параметров контекста."""

        from m3.actions.context import (
            DeclarativeActionContext,
        )

        params = (
            (
                'm3-gar:unicode-or-none',
                lambda s: str(s) if s else None
            ),
            (
                'm3-gar:int-list',
                lambda s: [int(x) for x in s.split(',')]
            ),
            (
                'm3-gar:guid-or-none',
                lambda x: UUID(x) if x else None
            ),
        )

        for name, parser in params:
            DeclarativeActionContext.register_parser(name, parser)

    def init(self):
        """Регистрирует наборы действий в M3."""

        from m3_gar_client import (
            config,
        )
        from m3_gar_client.backends.m3_rest_gar.proxy.actions import (
            Pack,
        )

        self._register_parsers()

        self._pack = Pack()

        config.controller.extend_packs((
            self._pack,
        ))

    def place_search_url(self):
        """URL для поиска населенных пунктов."""

        return self._pack.place_search_action.get_absolute_url()

    def street_search_url(self):
        """URL для поиска улиц."""

        return self._pack.street_search_action.get_absolute_url()

    def house_search_url(self):
        """URL для запроса списка домов."""

        return self._pack.house_search_action.get_absolute_url()



import warnings

from django.apps.config import (
    AppConfig as AppConfigBase,
)

import m3_gar_client


class AppConfig(AppConfigBase):

    name = __package__

    def ready(self):
        super().ready()

        if m3_gar_client.config:
            m3_gar_client.config.backend.init()
        else:
            warnings.warn("Не указана конфигурация m3-gar, пакет не будет инициализирован")



from django.core.cache import (
    cache,
)
from django.core.management import (
    BaseCommand,
)

from m3_gar_client.backends.m3_rest_gar.server import (
    server,
)


class Command(BaseCommand):
    """Команда очищает все ключи кеша ГАР"""

    def handle(self, *args, **options):
        cache.delete_many(keys=cache.keys(f'{server.cache_key_prefix}*'))
        self.stdout.write('Очистка завершена')



from django.apps import (
    AppConfig as AppConfigBase,
)


class AppConfig(AppConfigBase):

    name = __package__

    def _init_m3_gar_client(self):
        """Настраивает django-приложение m3_gar_client."""
        import m3_gar_client
        m3_gar_client.config = m3_gar_client.Config()

    def ready(self):
        self._init_m3_gar_client()

        super().ready()



import http.client as http_client
from abc import (
    ABC,
    abstractmethod,
)
from itertools import (
    chain,
    count,
)
from typing import (
    Any,
    Generator,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Union,
)

from m3_gar_constants import (
    DEFAULT_PAGE_LIMIT,
    GAR_HIERARCHY_MUN,
    GAR_LEVELS_PLACE,
    GAR_LEVELS_STREET,
    RESULT_PAGE_SIZE as DEFAULT_RESULT_PAGE_SIZE,
)

from m3_gar_client.backends.m3_rest_gar.server import (
    server,
)
from m3_gar_client.data import (
    AddressObject,
    Apartment,
    House,
    ObjectMapper,
    Stead,
)
from m3_gar_client.utils import (
    cached_property,
)


# -----------------------------------------------------------------------------
# Обёртки над данными, поступающими с сервера m3-rest-gar.

class AddressObjectMapper(ObjectMapper):
    """Обертка над данными сервера m3-rest-gar об адресных объектах.

    Предназначена для использования при создании экземпляров.
    """

    fields_map = {
        'id': 'id',
        'guid': 'objectguid',
        'obj_id': 'objectid',
        'previous_id': 'previd',
        'next_id': 'nextid',
        'ifns_fl_code': 'params__IFNSFL__value',
        'ifns_fl_territorial_district_code': 'params__territorialifnsflcode__value',
        'ifns_ul_code': 'params__IFNSUL__value',
        'ifns_ul_territorial_district_code': 'params__territorialifnsulcode__value',
        'okato': 'params__OKATO__value',
        'oktmo': 'params__OKTMO__value',
        'postal_code': 'params__PostIndex__value',
        'formal_name': 'name',
        'official_name': 'name',
        'short_name': 'typename',
        'type_full_name': 'type_full_name',
        'level': 'level',
        'kladr_code': 'params__CODE__value',
        'kladr_plain_code': 'params__PLAINCODE__value',
        'date_of_update': 'updatedate',
        'date_of_creation': 'startdate',
        'date_of_expiration': 'enddate',
        'full_name': 'hierarchy__mun__name_with_parents',
        'has_children': 'hierarchy__mun__has_children',
        'region_code': 'region_code',
        'adm_parent_obj_id': 'hierarchy__adm__parentobjid__objectid',
        'adm_parent_guid': 'hierarchy__adm__parentobjid__objectguid',
        'mun_parent_obj_id': 'hierarchy__mun__parentobjid__objectid',
        'mun_parent_guid': 'hierarchy__mun__parentobjid__objectguid',
    }

    assert set(fields_map) == set(AddressObject.fields)


class HouseMapper(ObjectMapper):
    """Обертка над данными сервера m3-rest-gar об адресных объектах.

    Предназначена для использования при создании экземпляров.
    """

    fields_map = {
        'id': 'id',
        'guid': 'objectguid',
        'obj_id': 'objectid',
        'ifns_fl_code': 'params__IFNSFL__value',
        'ifns_fl_territorial_district_code': 'params__territorialifnsflcode__value',
        'ifns_ul_code': 'params__IFNSUL__value',
        'ifns_ul_territorial_district_code': 'params__territorialifnsulcode__value',
        'okato': 'params__OKATO__value',
        'oktmo': 'params__OKTMO__value',
        'postal_code': 'params__PostIndex__value',
        'house_number': 'housenum',
        'building_number': 'addnum1',
        'structure_number': 'addnum2',
        'date_of_update': 'updatedate',
        'date_of_creation': 'startdate',
        'date_of_end': 'enddate',
        'region_code': 'region_code',
        'adm_parent_obj_id': 'hierarchy__adm__parentobjid__objectid',
        'adm_parent_guid': 'hierarchy__adm__parentobjid__objectguid',
        'mun_parent_obj_id': 'hierarchy__mun__parentobjid__objectid',
        'mun_parent_guid': 'hierarchy__mun__parentobjid__objectguid',
        'level': 'hierarchy__mun__objectid__levelid',
        'number': 'number',
        'house_type': 'housetype__shortname',
        'house_type_full': 'housetype__name',
        'building_type': 'addtype1__shortname',
        'building_type_full': 'addtype1__name',
        'structure_type': 'addtype2__shortname',
        'structure_type_full': 'addtype2__name',
    }

    assert set(fields_map) == set(House.fields)


class SteadMapper(ObjectMapper):
    """Обертка над данными сервера m3-rest-gar об адресных объектах."""

    fields_map = {
        'id': 'id',
        'guid': 'objectguid',
        'obj_id': 'objectid',
        'ifns_fl_code': 'params__IFNSFL__value',
        'ifns_fl_territorial_district_code': 'params__territorialifnsflcode__value',
        'ifns_ul_code': 'params__IFNSUL__value',
        'ifns_ul_territorial_district_code': 'params__territorialifnsulcode__value',
        'okato': 'params__OKATO__value',
        'oktmo': 'params__OKTMO__value',
        'postal_code': 'params__PostIndex__value',
        'house_number': 'number',
        'date_of_update': 'updatedate',
        'date_of_creation': 'startdate',
        'date_of_end': 'enddate',
        'region_code': 'region_code',
        'adm_parent_obj_id': 'hierarchy__adm__parentobjid__objectid',
        'adm_parent_guid': 'hierarchy__adm__parentobjid__objectguid',
        'mun_parent_obj_id': 'hierarchy__mun__parentobjid__objectid',
        'mun_parent_guid': 'hierarchy__mun__parentobjid__objectguid',
        'level': 'hierarchy__mun__objectid__levelid',
        'number': 'number',
    }


class ApartmentMapper(ObjectMapper):
    """Обертка над данными сервера m3-rest-gar о помещениях."""

    fields_map = {
        'id': 'id',
        'guid': 'objectguid',
        'obj_id': 'objectid',
        'number': 'number',
        'apart_type_full': 'aparttype__name',
        'apart_type': 'aparttype__shortname',
        'adm_parent_obj_id': 'hierarchy__adm__parentobjid__objectid',
        'adm_parent_guid': 'hierarchy__adm__parentobjid__objectguid',
        'mun_parent_obj_id': 'hierarchy__mun__parentobjid__objectid',
        'mun_parent_guid': 'hierarchy__mun__parentobjid__objectguid',
        'level': 'hierarchy__mun__objectid__levelid',
    }


class UIAddressObjectMapper(ObjectMapper):
    """Обертка над данными сервера m3-rest-gar об адресных объектах.

    Предназначена для использования в UI.
    """

    fields_map = {
        'objectId': 'objectguid',
        'level': 'level',
        'shortName': 'typename',
        'typeFullName': 'type_full_name',
        'formalName': 'name',
        'postalCode': 'params__PostIndex__value',
        'fullName': 'hierarchy__mun__name_with_parents',    # оставлено для совместимости
        'displayName': 'hierarchy__mun__name_with_parents',
        'hasChildren': 'hierarchy__mun__has_children',
    }


class UIStreetMapper(ObjectMapper):
    """Обертка над данными сервера m3-rest-gar об улицах.

    Предназначена для использования в UI.
    """

    fields_map = {
        'objectId': 'objectguid',
        'level': 'level',
        'shortName': 'typename',
        'typeFullName': 'type_full_name',
        'formalName': 'name',
        'postalCode': 'params__PostIndex__value',
        'displayName': 'name_with_typename',
    }


class UIHouseMapper(ObjectMapper):
    """Обертка над данными сервера m3-rest-gar о зданиях.

    Предназначена для использования в UI.
    """

    @staticmethod
    def _get_display_name(drf_object_data):
        type_number_mapping = {
            'housetype__shortname': 'housenum',
            'addtype1__shortname': 'addnum1',
            'addtype2__shortname': 'addnum2',
        }
        parts = (
            f'{drf_object_data.get(k)} {drf_object_data.get(v)}'
            for k, v in type_number_mapping.items()
            if all((drf_object_data.get(k), drf_object_data.get(v)))
        )
        return ', '.join(parts)

    fields_map = {
        'objectId': 'objectguid',
        'houseNumber': 'housenum',
        'buildingNumber': 'addnum1',
        'structureNumber': 'addnum2',
        'postalCode': 'params__PostIndex__value',
        'houseType': 'housetype__shortname',
        'buildingType': 'addtype1__shortname',
        'structureType': 'addtype2__shortname',
        'displayName': _get_display_name,
    }


class UISteadMapper(ObjectMapper):
    """Обертка над данными сервера m3-rest-gar о земельных участках.

    Предназначена для использования в UI.
    """

    fields_map = {
        'objectId': 'objectguid',
        'steadNumber': 'number',
        'postalCode': 'params__PostIndex__value',
    }


class UIApartmentMapper(ObjectMapper):
    """Обертка над данными сервера m3-rest-gar о помещениях.

    Предназначена для использования в UI.
    """

    @staticmethod
    def _get_display_name(drf_object_data):
        apartment_number = drf_object_data.get('number')
        apart_type = drf_object_data.get('aparttype__shortname')

        if apart_type:
            full_name = f'{apart_type} {apartment_number}'
        else:
            full_name = apartment_number

        return full_name

    fields_map = {
        'objectId': 'objectguid',
        'number': 'number',
        'apartType': 'aparttype__shortname',
        'displayName': _get_display_name,
    }


# -----------------------------------------------------------------------------
# Загрузчики данных с сервера m3-rest-gar.

class LoaderResult(NamedTuple):
    """Результат загрузки данных."""

    rows: List[dict]
    total: int


class LoaderBase(ABC):
    """Базовый класс для загрузчиков объектов ГАР.

    Attributes:
        filter_string: Строка для фильтрации объектов.
        timeout: Timeout запросов к серверу ГАР в секундах.
    """

    def __init__(self, filter_string, **kwargs):
        self.filter_string = filter_string
        self.timeout = kwargs.get('timeout')
        self.exact = kwargs.get('exact', False)

    @property
    @abstractmethod
    def _path(self):
        """Путь к ресурсу API сервера ГАР."""

    @cached_property
    def _fields(self):
        """Имена полей, подлежащих загрузке."""

        return list(self._mapper_class.fields_map.keys())

    @property
    @abstractmethod
    def _mapper_class(self):
        """Класс, преобразующий имена полей."""

    @property
    def _filter_param(self):
        """Query-параметр для фильтрации."""

        return 'name'

    def _load_page(self, params, page):
        params = params.copy()
        params['page'] = page

        drf_response = server.get(self._path, params, timeout=self.timeout)
        if drf_response and drf_response.status_code == http_client.OK:
            result = drf_response.json()
        else:
            result = None

        return result

    def _process_object_data(self, drf_object_data):
        """Выполняет дополнительную обработку данных объекта ГАР.

        Args:
            drf_object_data: Данные объекта ГАР, полученные с сервера m3-rest-gar.
        """
        return map_object_data(self._mapper_class, drf_object_data)

    def _build_result(self, object_data):
        """Формирует данные результирующего объекта ГАР.

        Полученные данные включаются в результат загрузки.

        Args:
            object_data: Данные объекта, полученные с сервера ГАР и прошедшие обработку в методе
                         ``_process_object_data``.
        """

        return {field: object_data[field] for field in self._fields}

    @abstractmethod
    def _filter(self, object_data):
        """Возвращает True, если объект ГАР должен попасть в загрузку."""

    def _get_params(self):
        """Возвращает параметры запроса к серверу ГАР."""

        filter_param = self._filter_param

        if self.exact:
            filter_param += '__exact'

        result = {
            filter_param: self.filter_string,
        }

        return result

    def load_raw_page(self, page_number: int, params: Optional[dict] = None) -> Tuple[Generator, bool, int]:
        """Возвращает данные конкретной страницы адресных объектов в исходном виде.

        Также возвращает признак наличия следующей страницы с данными и общее число найденных объектов.
        """
        if params is None:
            params = self._get_params()

        results = ()
        total_count = 0
        has_next_page = False

        drf_data = self._load_page(params, page_number)

        if drf_data:
            results = drf_data.get('results', [])
            total_count = drf_data.get('count', len(results))
            has_next_page = drf_data.get('next') or False

        result_data = (
            object_data
            for object_data in (self._process_object_data(drf_data) for drf_data in results)
            if self._filter(object_data)
        )

        return result_data, has_next_page, total_count

    def load_raw(self, page: Optional[int] = None) -> Generator:
        """Возвращает данные адресных объектов в исходном виде.

        Args:
            page: Номера страницы для загрузки данных. ``None`` указывает на необходимость загрузки всех страниц.
        """

        from django.conf import (
            settings,
        )

        page_limit = settings.GAR.get('PAGE_LIMIT', DEFAULT_PAGE_LIMIT)

        if page is None:
            pages = count(start=1)
        else:
            pages = range(page, page + 1)

        params = self._get_params()

        for page_number in pages:
            result_data, has_next_page, total_count = self.load_raw_page(page_number, params)

            yield from result_data

            if not has_next_page or page_number >= page_limit:
                break

    def load_page_results(self, page_number: int) -> Tuple[Generator, int]:
        """Возвращает данные в соответствии с параметрами загрузчика по указанной странице."""
        result_data, _, total_count = self.load_raw_page(page_number)

        return (self._build_result(data) for data in result_data), total_count

    def load_results(self, page=None):
        """Возвращает данные в соответствии с параметрами загрузчика.

        Args:
            page: Номера страницы для загрузки данных. ``None`` указывает на необходимость загрузки всех страниц.
        """

        return map(self._build_result, self.load_raw(page))

    @abstractmethod
    def _sort_key(self, object_data):
        """Возвращает значение ключа для сортировки результатов загрузки.

        Args:
            object_data: Данные загруженного объекта ГАР.
        """

    def _process_result(self, data):
        """Обработка полученных после загрузки данных.

        Args:
            data: Кортеж словарей с данными загруженных объектов ГАР.
        """

        return sorted(data, key=self._sort_key)

    def load(self, page: Optional[int] = None) -> LoaderResult:
        """Загружает данные с сервера ГАР.

        Args:
            page: Номера страницы для загрузки данных. ``None`` указывает на необходимость загрузки всех страниц.
        """
        if page is None:
            data = self.load_results()
            total = None
        else:
            data, total = self.load_page_results(page)

        rows = self._process_result(data)
        total = total or len(rows)

        return LoaderResult(rows=rows, total=total)


class ParentFilterMixin:
    """Миксин для загрузчиков, которые используют id родительского объекта в запросах к серверу ГАР."""

    @property
    @abstractmethod
    def _default_hierarchy(self):
        """Вид иерархии, используемый по умолчанию."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._hierarchy = kwargs.get('hierarchy') or self._default_hierarchy
        self._parent_id = kwargs.get('parent_id')

    def _get_params(self):
        result = super()._get_params()

        if self._parent_id:
            result['parent'] = str('{}:{}'.format(self._hierarchy, self._parent_id))

        return result


class AddressObjectLoaderBase(ParentFilterMixin, LoaderBase):
    """Базовый класс для загрузчиков адресных объектов ГАР.

    В терминологии ГАР адресными объектами называются:

        * субъекты РФ;
        * административные и муниципальные районы субъектов РФ;
        * города;
        * сельские/городские поселения;
        * населенные пункты;
        * элементы планировочной структуры;
        * улицы.
    """

    _default_hierarchy = GAR_HIERARCHY_MUN

    @property
    def _path(self):
        return '/addrobj/'

    @property
    @abstractmethod
    def _levels(self):
        """Уровни адресных объектов, для которых нужно искать данные в ГАР."""

    def _get_params(self):
        """Возвращает параметры запроса к серверу ГАР."""

        result = super()._get_params()

        if self._levels:
            result['level'] = ','.join(map(str, self._levels))

        return result

    def _filter(self, object_data):
        return True

    def _sort_key(self, object_data):
        return object_data['fullName'] or ''


class AddressObjectLoader(AddressObjectLoaderBase):
    """Загрузчик адресных объектов ГАР.

    Загружает информацию об адресных объектах, соответствующих строке
    фильтрации и находящихся на одном из указанных в параметрах экземпляра
    уровней иерархии адресных объектов.
    """

    _levels = None

    _mapper_class = AddressObjectMapper

    def __init__(self, filter_string, levels=None, typenames=None, **kwargs):
        """Инициализация экземпляра класса.

        Args:
            filter_string: Строка для фильтрации объектов.
            levels: Уровни адресных объектов.
        """

        super().__init__(filter_string, **kwargs)

        self._levels = levels
        self._typenames = typenames

    def _get_params(self):
        result = super()._get_params()

        if self._typenames:
            result['typename'] = ','.join(map(str, self._typenames))

        return result


class PlaceLoader(AddressObjectLoaderBase):
    """Загрузчик сведений о населенных пунктах."""

    _levels = GAR_LEVELS_PLACE

    _mapper_class = UIAddressObjectMapper

    @property
    def _filter_param(self):
        """Query-параметр для фильтрации."""

        return 'name_with_parents'

    def _get_params(self):
        """Возвращает параметры запроса к серверу ГАР."""

        filter_param = self._filter_param

        if self.exact:
            filter_param += '__exact'

        result = {
            filter_param: f'{self._default_hierarchy}:{self.filter_string}',
        }

        return result

    def _filter(self, object_data):
        return bool(object_data['fullName'])

    def _sort_key(
        self,
        object_data: dict[str, Any]
    ) -> Tuple[Union[int, str], ...]:
        """
        Возвращает ключ для сортировки результатов поиска адресных объектов.

        Метод поддерживает два режима сортировки:
         - по уровню объекта, затем по алфавиту.
         - по количеству совпадений запроса с частями адреса,
           по совпадению первых букв, затем по алфавиту.
        """

        from django.conf import (
            settings,
        )

        full_name = object_data.get('fullName', '') or ''

        if not settings.GAR.get('USE_IMPROVED_SORTING', False):
            return object_data.get('level', 0), full_name

        filter_lower = getattr(self, '_cached_filter_lower', None)
        if filter_lower is None:
            filter_lower = self.filter_string.lower()
            setattr(self, '_cached_filter_lower', filter_lower)

        if filter_lower and filter_lower not in full_name.lower():
            return 0, 0, full_name

        parts = full_name.split(',')
        match_count = 0

        for part in parts:
            part = part.strip().lower()
            if filter_lower in part:
                match_count += 1

        starts_with_same_letter = (
            1 if parts and
            filter_lower and
            parts[0] and
            parts[0][0].lower() == filter_lower[0].lower()
            else 0
        )

        return -match_count, -starts_with_same_letter, full_name

    def load(self, page: Optional[int] = None) -> LoaderResult:
        """Загружает и подготавливает данные."""

        from django.conf import (
            settings,
        )

        def get_safe_config(key, default):
            value = settings.GAR.get(key)

            return value if value and value > 0 else default

        server_page_size = get_safe_config('RESULT_PAGE_SIZE', DEFAULT_RESULT_PAGE_SIZE)
        max_pages = get_safe_config('PAGE_LIMIT', DEFAULT_PAGE_LIMIT)

        total_pages = 1
        max_items = server_page_size * max_pages
        needed_api_pages = (max_items + DEFAULT_RESULT_PAGE_SIZE - 1) // DEFAULT_RESULT_PAGE_SIZE

        params = self._get_params()
        all_data = []

        for page_num in range(1, needed_api_pages + 2):
            result_data, has_next_page, total_count = self.load_raw_page(page_num, params)

            for item_data in result_data:
                all_data.append(self._build_result(item_data))
                if len(all_data) >= max_items:
                    break

            if len(all_data) >= max_items or not has_next_page:
                break

        sorted_data = self._process_result(all_data)
        data_pages = (len(sorted_data) + server_page_size - 1) // server_page_size
        actual_pages = min(max_pages, data_pages)

        if page is None:
            return LoaderResult(rows=sorted_data, total=total_pages)

        total_pages = actual_pages * DEFAULT_RESULT_PAGE_SIZE

        if page > actual_pages:
            return LoaderResult(rows=[], total=total_pages)

        start = (page - 1) * server_page_size
        end = start + server_page_size
        page_data = sorted_data[start:end]

        return LoaderResult(rows=page_data, total=total_pages)


class StreetLoader(AddressObjectLoaderBase):
    """Загрузчик сведений об улицах."""

    _levels = GAR_LEVELS_STREET

    _mapper_class = UIAddressObjectMapper

    @property
    def _filter_param(self):
        """Query-параметр для фильтрации."""

        return 'name_with_typename'


class HouseLoader(ParentFilterMixin, LoaderBase):
    """Загрузчик сведений о зданиях."""

    _default_hierarchy = GAR_HIERARCHY_MUN

    _mapper_class = HouseMapper

    @property
    def _path(self):
        return '/houses/'

    @property
    def _filter_param(self):
        return 'housenum'

    def _filter(self, object_data):
        return True

    def _sort_key(self, object_data):
        return object_data['house_number'] or ''

    @staticmethod
    def _split_number(number):
        """Разделяет номер на целочисленную и буквенную части.

        Args:
            number: Номер дома/корпуса/строения.
        """

        int_part = ''.join(ch for ch in number if ch.isdigit())
        str_part = number[len(int_part):]

        return int(int_part) if int_part else -1, str_part


class SteadLoader(ParentFilterMixin, LoaderBase):
    """Загрузчик сведений о земельных участках."""

    _default_hierarchy = GAR_HIERARCHY_MUN

    _mapper_class = SteadMapper

    @property
    def _path(self):
        return '/steads/'

    @property
    def _filter_param(self):
        return 'number'

    def _filter(self, object_data):
        return True

    def _sort_key(self, object_data):
        return object_data.get('stead_number', object_data.get('number')) or ''

    @staticmethod
    def _split_number(number):
        """Разделяет номер на целочисленную и буквенную части.

        Args:
            number: Номер земельного участка.
        """

        int_part = ''.join(ch for ch in number if ch.isdigit())
        str_part = number[len(int_part):]

        return int(int_part) if int_part else -1, str_part


class ApartmentLoader(ParentFilterMixin, LoaderBase):
    """Загрузчик сведений о помещениях."""

    _default_hierarchy = GAR_HIERARCHY_MUN

    _mapper_class = ApartmentMapper

    @property
    def _path(self):
        return '/apartments/'

    @property
    def _filter_param(self):
        return 'number'

    def _filter(self, object_data):
        return True

    def _sort_key(self, object_data):
        return object_data['number'] or ''

    @staticmethod
    def _split_number(number):
        """Разделяет номер на целочисленную и буквенную части.

        Args:
            number: Номер земельного участка.
        """

        int_part = ''.join(ch for ch in number if ch.isdigit())
        str_part = number[len(int_part):]

        return int(int_part) if int_part else -1, str_part


class UIStreetLoader(StreetLoader):
    _mapper_class = UIStreetMapper

    def _sort_key(self, object_data):
        return object_data['displayName']


class UIHouseLoader(HouseLoader):
    _mapper_class = UIHouseMapper

    def _sort_key(self, object_data):
        return tuple(chain(*(
            self._split_number(object_data[number_type + 'Number'])
            for number_type in ('house', 'building', 'structure')
        )))

    def _filter(self, object_data):
        """Возвращает True для записей, соответствующих параметрам поиска.

        Запись считается соответствующей указанным при инициализации загрузчика
        параметрам поиска, если:

            * номер дома (если есть) в записи **начинается со строки**,
              указанной в аргументе ``filter_string``;
            * номер корпуса или строения (если номер дома отсутствует) в записи
              **начинается со строки**, указанной в аргументе
              ``filter_string``;
            * в аргументе ``filter_string`` конструктора класса было передано
              значение ``None``.
        """
        filter_string_lower = self.filter_string.lower()

        if self.filter_string is None:
            result = True
        else:
            house = (object_data.get('houseNumber') or '').lower()

            if house:
                result = house.startswith(filter_string_lower)
            else:
                building = (object_data.get('buildingNumber') or '').lower()
                structure = (object_data.get('structureNumber') or '').lower()
                result = (
                    (building and building.startswith(filter_string_lower)) or
                    (structure and structure.startswith(filter_string_lower))
                )

        return result

    def _process_object_data(self, drf_object_data):
        house_data = super()._process_object_data(drf_object_data)

        house_data['houseNumber'] = house_data['houseNumber'] or ''
        house_data['buildingNumber'] = house_data['buildingNumber'] or ''
        house_data['structureNumber'] = house_data['structureNumber'] or ''

        return house_data


class UISteadLoader(SteadLoader):
    _mapper_class = UISteadMapper

    def _sort_key(self, object_data):
        return tuple(chain(*(
            self._split_number(object_data[number_type + 'Number'])
            for number_type in ('stead',)
        )))

    def _filter(self, object_data):
        """Возвращает True для записей, соответствующих параметрам поиска.

        Запись считается соответствующей указанным при инициализации загрузчика
        параметрам поиска, если:

            * номер участка в записи **начинается со строки**,
              указанной в аргументе ``filter_string``;
            * в аргументе ``filter_string`` конструктора класса было передано
              значение ``None``.
        """

        filter_string_lower = self.filter_string.lower()

        if self.filter_string is None:
            result = True
        else:
            stead = (object_data.get('steadNumber') or '').lower()
            result = stead.startswith(filter_string_lower)

        return result

    def _process_object_data(self, drf_object_data):
        stead_data = super()._process_object_data(drf_object_data)

        stead_data['steadNumber'] = stead_data['steadNumber'] or ''

        return stead_data


class UIHouseOrSteadLoader:

    """Фасад для загрузки домов и участков."""

    def __init__(self, *args, **kwargs):
        self._house_loader = UIHouseLoader(*args, **kwargs)
        self._stead_loader = UISteadLoader(*args, **kwargs)

    def load(self, page: Optional[int] = None) -> LoaderResult:
        house_result = self._house_loader.load(page=page)
        stead_result = self._stead_loader.load(page=page)
        total = house_result.total + stead_result.total

        return LoaderResult(
            rows=list(chain(house_result.rows, stead_result.rows)),
            total=total,
        )


class UIApartmentLoader(ApartmentLoader):

    _mapper_class = UIApartmentMapper


# -----------------------------------------------------------------------------
# Функции для создания объектов m3-gar-client на основе данных m3-rest-gar.
def get_address_object(obj_id, timeout=None):
    """Возвращает адресный объект, загруженный с сервера ГАР.

    Если адресный объект не найден, возвращает ``None``.

    Args:
        obj_id: ObjectID/GUID адресного объекта ГАР.
        timeout: timeout запроса к серверу ГАР в секундах.

    Returns:
        Адресный объект

    Exceptions:
        requests.ConnectionError: если не удалось соединиться с сервером ГАР
    """
    address_object = None

    if obj_id:
        response = server.get('/addrobj/{}/'.format(obj_id), timeout=timeout)

        if response and response.status_code == http_client.OK:
            response_data = response.json()
            mapped_data = map_object_data(AddressObjectMapper, response_data)
            address_object = AddressObject(**mapped_data)

    return address_object


def find_address_objects(
    filter_string,
    levels=None,
    typenames=None,
    parent_id=None,
    timeout=None,
):
    """Возвращает адресные объекты, соответствующие параметрам поиска.

    Args:
        filter_string: Строка поиска.
        levels: Уровни адресных объектов, среди которых нужно осуществлять поиск.
        parent_id: ID родительского объекта.
        timeout: Timeout запросов к серверу ГАР в секундах.
    """

    return AddressObjectLoader(
        filter_string,
        levels=levels,
        typenames=typenames,
        parent_id=parent_id,
        timeout=timeout,
    ).load_results()


def get_stead(stead_id, timeout=None):
    """
    Возвращает информацию о земельном участке по его ID в ГАР.

    Args:
        stead_id: ID земельного участка.
        timeout: Timeout запросов к серверу ГАР в секундах.

    Returns:
        Объект m3_gar_client.data.Stead
    """

    assert stead_id is not None

    response = server.get('/steads/{}/'.format(stead_id), timeout=timeout)

    if response and response.status_code == http_client.OK:
        response_data = response.json()
        mapped_data = map_object_data(SteadMapper, response_data)
        stead = Stead(**mapped_data)
    else:
        stead = None

    return stead


def get_house(house_id, timeout=None):
    """Возвращает информацию о здании по его ID в ГАР.

    Args:
        house_id: ID здания.
        timeout: Timeout запросов к серверу ГАР в секундах.
    """

    assert house_id is not None

    response = server.get('/houses/{}/'.format(house_id), timeout=timeout)

    if response and response.status_code == http_client.OK:
        response_data = response.json()
        mapped_data = map_object_data(HouseMapper, response_data)
        result = House(**mapped_data)
    else:
        result = None

    return result


def find_house(house_number='', parent_id=None, building_number=None, structure_number=None, timeout=None):
    """Возвращает информацию о здании по его номеру.

    Args:
        house_number: Номер дома.
        parent_id: ID родительского объекта.
        building_number: Номер корпуса.
        structure_number: Номер строения.
        timeout: Timeout запросов к серверу ГАР в секундах.
    """

    result = HouseLoader(
        house_number,
        parent_id=parent_id,
        exact=True,
        timeout=timeout,
    ).load()

    houses = result.rows

    if len(houses) > 1:
        _filter = lambda house: (
            house['building_number'] == building_number and
            house['structure_number'] == structure_number
        )
        houses = list(filter(_filter, houses))

    if len(houses) == 1:
        house = House(**houses[0])
    else:
        house = None

    return house


def find_stead(number='', parent_id=None, timeout=None):
    """Возвращает информацию об участке по его номеру.

    :param str number: Номер участка.
    :param parent_id: ID родительского объекта.
    :param float timeout: Timeout запросов к серверу ГАР в секундах.

    :rtype: m3_gar_client.data.Stead or NoneType
    """
    result = SteadLoader(
        number,
        parent_id=parent_id,
        timeout=timeout,
    ).load()

    steads = result.rows

    if len(steads) == 1:
        stead = Stead(**steads[0])
    else:
        stead = None

    return stead


def get_apartment(apartment_id, timeout=None):
    """Возвращает информацию о помещении по его ID в ГАР.

    Args:
        apartment_id: ID помещения.
        timeout: Timeout запросов к серверу ГАР в секундах.
    """
    assert apartment_id is not None

    response = server.get('/apartments/{}/'.format(apartment_id), timeout=timeout)

    if response and response.status_code == http_client.OK:
        response_data = response.json()
        mapped_data = map_object_data(ApartmentMapper, response_data)
        apartment = Apartment(**mapped_data)
    else:
        apartment = None

    return apartment


def find_apartment(number='', parent_id=None, timeout=None):
    """Возвращает информацию о помещении по его номеру.

    Args:
        number: Номер квартиры.
        parent_id: ID родительского объекта.
        timeout: Timeout запросов к серверу ГАР в секундах.
    """

    result = ApartmentLoader(
        number,
        parent_id=parent_id,
        exact=True,
        timeout=timeout,
    ).load()

    apartments = result.rows

    if len(apartments) == 1:
        apartment = Apartment(**apartments[0])
    else:
        apartment = None

    return apartment


def flatten_object_data(raw_object_data):
    """Преобразует данные объекта ГАР к "плоскому" виду.

    Args:
        raw_object_data: Необработанные данные объекта ГАР, полученные с сервера m3-rest-gar.
    """

    flat_data = {}

    for key, value in raw_object_data.items():
        if type(value) is dict:
            for inner_key, inner_value in flatten_object_data(value).items():
                flat_data['{}__{}'.format(key, inner_key)] = inner_value

        elif type(value) is list and key == 'params':
            for param_dict in value:
                if type(param_dict) is not dict:
                    continue

                code = param_dict.get('typeid', {}).get('code')
                if code is None:
                    continue

                for inner_key, inner_value in flatten_object_data(param_dict).items():
                    flat_data['params__{}__{}'.format(code, inner_key)] = inner_value

        else:
            flat_data[key] = value

    return flat_data


def map_object_data(mapper_class, raw_object_data):
    """Выполняет маппинг сырых данных, полученных с сервера m3-rest-gar.

    Args:
        mapper_class: Класс-маппер.
        raw_object_data: Необработанные данные объекта ГАР, полученные с сервера m3-rest-gar.
    """

    flat_data = flatten_object_data(raw_object_data)
    mapped_data = mapper_class(flat_data)

    return mapped_data



from abc import (
    ABC,
    abstractmethod,
)


class BackendBase(ABC):

    """Базовый класс для бэкендов."""

    def init(self):
        """Метод инициализации бекенда и бутстреппинга пакета."""

    @property
    @abstractmethod
    def place_search_url(self):
        """URL для поиска населенных пунктов.

        :rtype: str
        """

    @property
    @abstractmethod
    def street_search_url(self):
        """URL для поиска улиц.

        :rtype: str
        """

    @property
    @abstractmethod
    def house_search_url(self):
        """URL для запроса списка домов.

        :rtype: str
        """

    def configure_place_field(self, field):
        """Настраивает поле "Населенный пункт".

        :param field: Поле "Населенный пункт".
        :type field: m3_ext.ui.fields.simple.ExtComboBox
        """

    def configure_street_field(self, field):
        """Настраивает поле "Улица".

        :param field: Поле "Улица".
        :type field: m3_ext.ui.fields.simple.ExtComboBox
        """

    def configure_house_field(self, field):
        """Настраивает поле "Дом".

        :param field: Поле "Дом".
        :type field: m3_ext.ui.fields.simple.ExtComboBox
        """

    @abstractmethod
    def find_address_objects(
        self,
        filter_string,
        levels=None,
        typenames=None,
        parent_id=None,
        timeout=None,
    ):
        """Возвращает адресные объекты, соответствующие параметрам поиска.

        :param str filter_string: Строка поиска.
        :param levels: Уровни адресных объектов, среди которых нужно осуществлять поиск.
        :param parent_id: ID родительского объекта.
        :param float timeout: Timeout запросов к серверу ГАР в секундах.

        :rtype: generator
        """

    @abstractmethod
    def get_address_object(self, obj_id, timeout=None):
        """Возвращает адресный объект ГАР по его ID.

        :param obj_id: ID адресного объекта ГАР.
        :param float timeout: Timeout запросов к серверу ГАР в секундах.

        :rtype: m3_gar_client.data.AddressObject
        """

    @abstractmethod
    def find_house(self, house_number, parent_id, building_number, structure_number, timeout=None):
        """Возвращает информацию о здании по его номеру.

        Args:
            house_number: Номер дома
            parent_id: ID родительского объекта
            building_number: Номер корпуса
            structure_number: Номер строения
            timeout: Timeout запросов к серверу ГАР в секундах

        Returns:
            Найденные объекты зданий
        """

    @abstractmethod
    def get_house(self, house_id, timeout=None):
        """Возвращает информацию о здании по его ID в ГАР.

        :param house_id: ID здания.
        :param float timeout: Timeout запросов к серверу ГАР в секундах.

        :rtype: m3_gar_client.data.House
        """

    @abstractmethod
    def find_stead(self, number, parent_id, timeout=None):
        """Возвращает информацию об участке по его номеру.

        Args:
            number: Номер участка
            parent_id: ID родительского объекта
            timeout: Timeout запросов к серверу ГАР в секундах

        Returns:
            Найденные земельные участки
        """

    @abstractmethod
    def get_stead(self, stead_id, timeout=None):
        """Возвращает информацию о земельном участке по его ID в ГАР.

        Args:
            stead_id: ID земельного участка.
            timeout: Timeout запросов к серверу ГАР в секундах.

        Returns:
            Объект m3_gar_client.data.House
        """

    @abstractmethod
    def find_apartment(self, number, parent_id, timeout=None):
        """Возвращает информацию о помещении по его номеру.

        Args:
            number: Номер квартиры.
            parent_id: ID родительского объекта.
            timeout: Timeout запросов к серверу ГАР в секундах.
        """

    @abstractmethod
    def get_apartment(self, apartment_id, timeout=None):
        """Возвращает информацию о помещении по его ID в ГАР.

        Args:
            apartment_id: ID помещения.
            timeout: Timeout запросов к серверу ГАР в секундах.

        Returns:
            Объект m3_gar_client.data.Apartment
        """



from abc import (
    ABCMeta,
    abstractmethod,
)
from itertools import (
    chain,
)
from typing import (
    Optional,
)

from django.http.response import (
    JsonResponse,
)
from m3.actions import (
    Action,
    ActionPack,
)

from m3_gar_client.backends.m3_rest_gar.utils import (
    PlaceLoader,
    StreetLoader,
    UIHouseLoader,
    UISteadLoader,
)
from m3_gar_client.utils import (
    correct_keyboard_layout,
)


class ActionBase(Action, metaclass=ABCMeta):
    """
    Базовый класс для обработчиков запросов на поиск данных в ГАР.
    """

    # Признак того, что необходимо корректировать пользовательский ввод
    correct_keyboard_input = True

    @abstractmethod
    def _get_loader(self, context):
        """Возвращает загрузчик данных.

        :rtype: m3_gar_client.backends.m3_rest_gar.utils.LoaderBase
        """

    def context_declaration(self):
        return {
            'filter': {
                'type': 'unicode',
                'default': '',
            },
        }

    @staticmethod
    def get_page_number(context) -> Optional[int]:
        """Возвращает номер страницы из параметров запроса для загрузки данных.

        В ExtJs для постраничного получения данных используются параметры start и limit,
        тогда как в m3_rest_gar сервисе используется постраничный запрос с параметром page.
        Необходимо из значений start и limit определить номер страницы page.
        """
        start = getattr(context, 'start', None)
        limit = getattr(context, 'limit', None)
        page = None

        if start and limit:
            try:
                page = int(int(start) / int(limit)) + 1
            except (ZeroDivisionError, ValueError):
                page = 1

        return page

    def run(self, request, context):
        if self.correct_keyboard_input:
            context.filter = correct_keyboard_layout(context.filter)

        loader = self._get_loader(context)
        result = loader.load(page=self.get_page_number(context))

        return JsonResponse({
            'rows': result.rows,
            'total': result.total,
        })


class PlaceSearchAction(ActionBase):
    """
    Обработчик запросов на поиск населенных пунктов.
    """

    url = '/search/place'

    def _get_loader(self, context):
        # pylint: disable=abstract-class-instantiated
        return PlaceLoader(context.filter)


class ParentMixin:

    def context_declaration(self):
        result = super().context_declaration()

        result.update({
            'parent': {
                'type': 'unicode',
            },
        })

        return result


class StreetSearchAction(ParentMixin, ActionBase):
    """
    Обработчик запросов на поиск улиц в населенном пункте.
    """

    url = '/search/street'

    def _get_loader(self, context):
        # pylint: disable=abstract-class-instantiated
        return StreetLoader(context.filter, parent_id=str(context.parent))


class HouseSearchAction(ParentMixin, ActionBase):
    """
    Обработчик запросов на поиск домов.
    """

    url = '/search/house'
    correct_keyboard_input = False

    def _get_loader(self, context):
        # pylint: disable=abstract-class-instantiated
        return UIHouseLoader(context.filter, parent_id=context.parent)


class HouseOrSteadSearchAction(ParentMixin, ActionBase):
    """
    Обработчик запросов на поиск домов и земельных участков.
    """

    url = '/search/house_or_stead'
    correct_keyboard_input = False

    def run(self, request, context):
        if self.correct_keyboard_input:
            context.filter = correct_keyboard_layout(context.filter)

        loader_of_houses = self._get_loader(context)
        loader_of_steads = self.get_steads(context)

        house_result = loader_of_houses.load()
        stead_result = loader_of_steads.load()

        return JsonResponse({
            'rows': list(chain(house_result.rows, stead_result.rows)),
            'total': house_result.total + stead_result.total,
        })

    def _get_loader(self, context):
        # pylint: disable=abstract-class-instantiated
        return UIHouseLoader(context.filter, parent_id=context.parent)

    def get_steads(self, context):
        # pylint: disable=abstract-class-instantiated
        return UISteadLoader(context.filter, parent_id=context.parent)


class Pack(ActionPack):
    """
    Набор действий для проксирования запросов к серверу ГАР.
    """

    url = '/gar'

    def __init__(self):
        super().__init__()

        self.place_search_action = PlaceSearchAction()
        self.street_search_action = StreetSearchAction()
        self.house_search_action = HouseOrSteadSearchAction()

        self.actions.extend((
            self.place_search_action,
            self.street_search_action,
            self.house_search_action,
        ))



from django.urls import (
    path,
)

from . import (
    const,
    views,
)


urlpatterns = [
    path(
        'search/place/', views.PlaceSearchView.as_view(),
        name=const.SEARCH_PLACE_NAME
    ),
    path(
        'search/street/', views.StreetSearchView.as_view(),
        name=const.SEARCH_STREET_NAME
    ),
    path(
        'search/house/', views.HouseSearchView.as_view(),
        name=const.SEARCH_HOUSE_NAME
    ),
    path(
        'search/house-or-stead/', views.HouseOrSteadSearchView.as_view(),
        name=const.SEARCH_HOUSE_OR_STEAD_NAME
    ),
    path(
        'search/apartment/', views.ApartmentSearchView.as_view(),
        name=const.SEARCH_APARTMENT
    ),
]



from django.contrib.auth.models import (
    User,
)
from django.test import (
    Client,
)
from django.test.testcases import (
    TestCase,
)

from m3_gar_client.backends.m3_rest_gar.proxy_rest import (
    Backend,
)


class BackendsResultsTestCase(TestCase):

    def test_backends_results(self):
        backend = Backend()

        client = Client()
        user, _ = User.objects.get_or_create(username='testuser')
        client.force_login(user)  # Запросы требуют авторизации

        # Проверка получения нас. пункта
        place_url = backend.place_search_url
        self.assertIsInstance(place_url, str)
        self.assertGreater(len(place_url), 0)

        response = client.post(place_url, {'filter': 'ujhjl Yjdjcb,bhcr'})
        self.assertGreater(response.json()['total'], 0)

        place = response.json()['rows'][1]

        # Проверка получения улицы в нас. пункте
        street_url = backend.street_search_url
        self.assertIsInstance(street_url, str)
        self.assertGreater(len(street_url), 0)

        response = client.post(street_url, {'filter': 'Ktybyf', 'parent': place['objectId']})
        self.assertGreater(response.json()['total'], 0)

        street = response.json()['rows'][1]

        # Проверка получения дома на улице
        house_url = backend.house_search_url
        self.assertIsInstance(house_url, str)
        self.assertGreater(len(house_url), 0)

        response = client.post(house_url, {'filter': '18', 'parent': street['objectId']})
        self.assertGreater(response.json()['total'], 0)
        house = response.json()['rows'][0]

        # проверка получения квартиры в доме
        apartment_url = backend.apartment_search_url
        self.assertIsInstance(apartment_url, str)
        self.assertGreater(len(apartment_url), 0)

        response = client.post(apartment_url, {'filter': '84', 'parent': house['objectId']})
        self.assertGreater(response.json()['total'], 0)



from rest_framework.authentication import (
    SessionAuthentication,
)


class CsrfExemptSessionAuthentication(SessionAuthentication):

    """Авторизация без проверки CSRF токена.

    Пользователь должен быть авторизован, но проверка CSRF токена не
    выполняется, т.к. токен не передается с формы.
    """

    def enforce_csrf(self, request):
        return
